;*****************************************************************************
;  FILENAME: SDCard.asm
;  Version: 1.1
;  DESCRIPTION: SDCard User Module software implementation file
;
;  (c) Ilya Mamontov 2008-2009.
;
;
;	Just write code of SDCard_Callback_Procedure
;   and call SDCard_Run routine!
;
;            32 .
;        LCD ( 16  ).
;       - ""         (      ).
;   LCD    P2 ( )
;   PSOC +5V, SD +3V (  PSOC -  )
;*****************************************************************************

include "m8c.inc"
include "memory.inc"
include "SDCard.inc"


;-----------------------------------------------
;  Macros Definitions
;-----------------------------------------------
macro SDCard_DelayMin ;a minimum setup time for the high/low transition
	; 3 MHz: 1 nop operation
	; 6 MHz: 2 of nop
	;12 MHz: 3 of nop
	;24 MHz: 6 of nop
	nop	
	nop	
	nop	
	nop	
	nop	
	nop	
endm
;-----------------------------------------------
;	Drive the SD Card pins
macro SDCard_CS0
	and reg[SDCard_CS_Port],~SDCard_CS_Mask	;Drive the CS pin low
	nop
	SDCard_DelayMin							;Wait a minimum setup time for the high/low transition
endm
;-----------------------------------------------
macro SDCard_CS1
	or reg[SDCard_CS_Port],SDCard_CS_Mask	;Drive the CS pin high
	nop
	SDCard_DelayMin							;Wait a minimum setup time for the high/low transition
endm
;-----------------------------------------------
macro SDCard_DI0
	and reg[SDCard_DI_Port],~SDCard_DI_Mask	;Drive the DI pin low
endm
;-----------------------------------------------
macro SDCard_DI1
	or reg[SDCard_DI_Port],SDCard_DI_Mask	;Drive the DI pin high
endm
;-----------------------------------------------
macro SDCard_CLK0
	and reg[SDCard_CLK_Port],~SDCard_CLK_Mask	;Drive the CLK pin low
endm
;-----------------------------------------------
macro SDCard_CLK1
	or reg[SDCard_CLK_Port],SDCard_CLK_Mask	;Drive the CLK pin high
endm
;-----------------------------------------------
macro SDCard_Cmd	;@0, @1, @2, @3, @4, @5
	mov A,0xFF						;Send an 0xFF to insure data line is high before command sent
	call SDCard_SendByte
	mov A,0xFF
	call SDCard_SendByte
	SDCard_CS0						;Set SD Card select line low 

	mov A,@0
	call SDCard_SendByte
	mov A,@1
	call SDCard_SendByte
	mov A,@2
	call SDCard_SendByte
	mov A,@3
	call SDCard_SendByte
	mov A,@4
	call SDCard_SendByte
	mov A,@5
	call SDCard_SendByte
endm
;-----------------------------------------------
macro SDCard_CmdA	;@0, @1
	mov A,0xFF						;Send an 0xFF to insure data line is high before command sent
	call SDCard_SendByte
	mov A,0xFF
	call SDCard_SendByte
	SDCard_CS0						;Set SD Card select line low 

	mov A,@0
	call SDCard_SendByte
	mov A,[@1+3]
	call SDCard_SendByte
	mov A,[@1+2]
	call SDCard_SendByte
	mov A,[@1+1]
	call SDCard_SendByte
	mov A,[@1+0]
	call SDCard_SendByte
	mov A,0xFF
	call SDCard_SendByte
endm
;-----------------------------------------------
macro SDCard_Cmd_00		;Sends the six byte reset command with 7 bit CRC of 0x95 (Byte String = 0x40 0x00 0x00 0x00 0x00 0x95) 
	mov A,0xFF						;Send a null byte to shift for byte read
	call SDCard_SendByte
	mov A,0xFF
	call SDCard_SendByte
	call SDCard_Wait1ms				;Wait a minimum setup time for the high/low transition
	SDCard_CS0						;Set SD Card select line low 

	mov A,0x40
	call SDCard_SendByte
	mov A,0
	call SDCard_SendByte
	mov A,0
	call SDCard_SendByte
	mov A,0
	call SDCard_SendByte
	mov A,0
	call SDCard_SendByte
	mov A,0x95
	call SDCard_SendByte
endm
;-----------------------------------------------
macro SDCard_CmdEnd
	mov A,0xFF						;Send a null byte (Min 8 clock cycles required)	
	call SDCard_SendByte
	SDCard_CS1						;Set SD Card select line high (disables card for SPI)
endm
;-----------------------------------------------
macro SDCard_GetR1	;Gets R1 command response:
					;Status bits 0=Idle, 1=Erase Reset, 2=Illegal CMD, 3=CRC Error, 4=Erase_Seq_Error, 5=Address Error, 6=Param Error
	mov A,0xFF
	call SDCard_SendByte
	call SDCard_ReadByte	;result in A
endm
;-----------------------------------------------
macro SDCard_GetR2	;Gets R2 command response
	mov A,0xFF
	call SDCard_SendByte
	call SDCard_ReadByte
	call SDCard_ReadByte	;result in A
endm
;-----------------------------------------------
macro SDCard_SkipM	@0	;Read @0 dummy bytes
	mov [cnt],@0
	call SDCard_Skip
endm
;-----------------------------------------------
macro SDCard_ReadBuffM	@0,@1	;Read @0 bytes into buffer @1
	mov [cnt],@0
	mov X,@1
	call SDCard_ReadBuff
endm
;-----------------------------------------------
macro SDCard_Move	@0,@1	;Copy @0 <- @1
	mov [@0+0],[@1+0]
	mov [@0+1],[@1+1]
	mov [@0+2],[@1+2]
	mov [@0+3],[@1+3]
endm
;-----------------------------------------------
macro SDCard_Add	@0,@1	;@0 = @0 + @1
	mov A,[@1+0]					;DataA =DataA + DataAddr 
	add   [@0+0],A
	mov A,[@1+1]
	adc   [@0+1],A
	mov A,[@1+2]
	adc   [@0+2],A
	mov A,[@1+3]
	adc   [@0+3],A
endm
;-----------------------------------------------
macro SDCard_mul @0,@1,@2	;Unsigned multiplication @0(4) = @1(2) * @2(1)
	mov [@0+0],0	;clear  result
	mov [@0+1],0
	mov [@0+2],0
	mov [@0+3],0

	mov [tmp],[@2]	;copy @2 -> temp
	mov [cnt],8		;load count
.mul_1:
	asl [@0+0]		;shift left @0
	rlc [@0+1]
	rlc [@0+2]
	rlc [@0+3]

	rlc [tmp]		;shift left tmp (@2)
	jnc .mul_2		;C=0

	mov A,[@1+0]	;else @0 = @0 + @1
	add   [@0+0],A
	mov A,[@1+1]
	adc   [@0+1],A
	adc   [@0+2],0
	adc   [@0+3],0
.mul_2:
	dec [cnt]			
	jnz .mul_1		;repeat till 8 bits are done
endm
;-----------------------------------------------



;-----------------------------------------------
; Variable Allocation
;-----------------------------------------------
area SDCard_RAM (RAM, REL, CON)
;------------ work variables ------------
tmp:			BLK	2
cnt:			BLK	1	;counter of bytes for SPI
cnt8:			BLK	1	;counter of bits for SPI
cntMax:			BLK	2	;counter of dir entries
ptr::			BLK	2	;pointer to current char     (for _FindFile procedures)
ptrA:			BLK	1 	;pointer to alternative char (for _FindFile procedures) - if used!
altName::		BLK	4 	;(max) alternative char array(for _FindFile procedures) - if used!

buff::			BLK	BUFFERSIZE*BUFFNUMBER 	;data buffer (size is defind in SDCard.inc)
ptrR::			BLK	1 	;pointer to read from buffer
ptrW::			BLK	1 	;pointer to write into buffer
;---------- work FAT16 variables ----------
FClst::			BLK	2	;File start cluster
FSize::			BLK	4	;File size (long)
DataA:			BLK	4	;Address of current DATA block
ClstN::			BLK	2	;Current cluster number

;------------ FAT16 parameters ------------
FatAddr::				BLK	4	;Address of FAT area
DirAddr::				BLK	4	;Address of DIR area
DataAddr::				BLK	4	;Address of DATA area
Sectors_per_cluster:	BLK	1	;Number of sectors per cluster 1...64 (Usually one)
N_of_dir_enries:		BLK	2	;Number of root directory enries      (Usually 512)
FSizeCnt::						;(4) work variable
N_of_FAT_tables:		BLK	1	;Number of FAT tables                 (Standard 2)
N_of_sectors_per_FAT:	BLK	3	;(2) Number of sectors per FAT 1..64K (Depends on disk size)

area UserModules (ROM, REL)

;-----------------------------------------------------------------------------
;  FUNCTION: SDCard_SendByte
;       Sends a byte (using SPI protocol)
;  ARGUMENTS:
;     A - data to transmit.
;  RETURNS:  none
;-----------------------------------------------------------------------------
.SECTION
SDCard_SendByte:
  	mov [cnt8],8
.lp8:
	rlc A
	jc .c1
		SDCard_DI0
		SDCard_CLK0
		jmp .cnext
	.c1:SDCard_DI1
		SDCard_CLK0
		jmp .cnext
	.cnext:
		nop
	SDCard_CLK1
	dec [cnt8]
	jnz .lp8
	SDCard_CLK0
  	ret
.ENDSECTION

;-----------------------------------------------------------------------------
;  FUNCTION: SDCard_ReadByte
;       Receives a byte (using SPI protocol)
;  ARGUMENTS: none
;  RETURNS:  A - received data
;-----------------------------------------------------------------------------
.SECTION
SDCard_ReadByte:
  	mov [cnt8],8
.loop8:
	SDCard_CLK1
	tst reg[SDCard_DO_Port],SDCard_DO_Mask	;Check the DO pin
	jz .z0
		or  F,4		;C=1
		jmp .znext
	.z0:and F,~4	;C=0
		jmp .znext
	.znext:
	rlc A
	SDCard_CLK0
	dec [cnt8]
	jnz .loop8
  	ret
.ENDSECTION

;-----------------------------------------------------------------------------
;  FUNCTION: SDCard_Skip
;       Reads N dummy bytes
;  ARGUMENTS:
;     [cnt] - number of bytes to read 1...255 (0=256)
;  RETURNS:  none
;-----------------------------------------------------------------------------
.SECTION
SDCard_Skip:
.loop:
	call SDCard_ReadByte
	dec [cnt]
	jnz .loop
	ret
.ENDSECTION

;-----------------------------------------------------------------------------
;  FUNCTION: SDCard_ReadBuff
;       Reads N bytes from SD Card to buffer [X]
;  ARGUMENTS:
;     X - address of buffer
;     [cnt] - number of bytes
;  RETURNS:  none
;-----------------------------------------------------------------------------
.SECTION
SDCard_ReadBuff:
.loop:
	call SDCard_ReadByte
	mov [X],A
	inc X
	dec [cnt]
	jnz .loop
	ret
.ENDSECTION

;-----------------------------------------------------------------------------
;  FUNCTION: SDCard_CheckReply
;       Waits for reply byte 0xFE
;  ARGUMENTS:
;     [Reply_byte] (1)
;  RETURNS:  zero (Z=1) if success
;            nz   (Z=0) if card failed
;-----------------------------------------------------------------------------
.SECTION
SDCard_CheckReplyFE:
	mov [cnt],254				;Set number of tries until you timeout
.wait:
	call SDCard_ReadByte
	cmp A,0xFE
	jz .endwait					;Loop until status is good, or
	dec [cnt]
	jnc .wait					;until a timeout occurs 
.endwait:
	ret
.ENDSECTION

;-----------------------------------------------------------------------------
;  FUNCTION: SDCard_Wait1ms
;       Delay 1ms for any clock frequency
;       A is modified!
;-----------------------------------------------------------------------------
.SECTION
SDCard_Wait1ms:
	M8C_SetBank1
	mov A,reg[OSC_CR0]			;get CPU clock
	M8C_SetBank0
	and A,07h 					;mask off only the clock bits
	index SDCard_Delay_Table	;get delay value
	mov [cnt],A
	.loop:				;x 10 us
		nop
		mov [cnt8],18
		.loop10us:
			dec [cnt8]
		jnz .loop10us
	dec [cnt]
	jnz .loop
  	ret
.ENDSECTION
.LITERAL
SDCard_Delay_Table: DB   13,  25,   50,  100,    7,     4,    2,    1
;---------------------- 3MHz 6MHz 12MHz 24MHz 1.5MHz 750kHz 188kHz 94kHz
.ENDLITERAL


;-----------------------------------------------------------------------------
;  FUNCTION: SDCard_InitCard
;       Initializes SD card and gets FAT16 information
;  RETURNS:
;    A: 0=Success, 1=SD card failed, 2=FAT type is missing
;         Array of FAT16 parameters:
;    [FatAddr]				(4) - Address of FAT area
;    [DirAddr]				(4) - Address of DIR area
;    [DataAddr]				(4) - Address of DATA area
;    [Sectors_per_cluster]	(1) - Number of sectors per cluster 1...64 (Usually one)
;    [N_of_dir_enries]		(2) - Number of root directory enries     (Usually 512)
;    [N_of_FAT_tables]		(1) - Number of FAT tables               (Standard 2)
;    [N_of_sectors_per_FAT]	(2) - Number of sectors per FAT 1..64K  (Depends on disk size)
;-----------------------------------------------------------------------------
.SECTION
 SDCard_InitCard::
_SDCard_InitCard::
	SDCard_CS1
    ;If CPU clock = 24 MHz then set 12 MHz for use under 400Khz (Must run slowly until card initialized as SD or MMC)
    M8C_SetBank1
	mov A,reg[OSC_CR0]
	mov [tmp],A				;save current value
	and A,0x07				;mask off only the clock bits
	cmp A,3					;clock=24 MHz?
	jnz skip_12MHz			;no
		and reg[OSC_CR0],~1	;clear bit 0 (24 -> 12 MHz)
	skip_12MHz:
    M8C_SetBank0
   	;Wait at least 1 ms after power up before attempting to do anything with the card
	call SDCard_Wait1ms
	;Send out a string of at least 74 clocked highs (10 or more 0xFF characters) previous to sending commands
	mov [cnt],10
	.loop:
		mov A,0xFF ;FF
		call SDCard_SendByte
		dec [cnt]
	jnz .loop  
	;Set card to SPI mode - Set SD Card pin 1 low and then send CMD(0) Reset
	;If the CS line is low then a mode R1 SPI response is generated.
	SDCard_Cmd_00
	;;;;;;;; 1- ! ;;;;;;;;;;
continue1:
	SDCard_GetR1
	SDCard_CmdEnd
	;Test for SD card comptibility - The SD Card must be polled with a ACMD(41) until it's init is finished
	mov [cnt],255					;Set number of tries until you timeout
.looptest:
	SDCard_Cmd 0x77,0,0,0,0,0xFF	;Send CMD55 command string (next command is an ACMD)
	SDCard_GetR1
	SDCard_CmdEnd
	SDCard_Cmd 0x69,0,0,0,0,0xFF	;Send CMD41 command string (REQUIRES a CMD55 before)
	SDCard_GetR1
	mov X,A	;save result
	SDCard_CmdEnd
	dec [cnt]
	jz noCARD						;Loop until a timeout occurs, or
	mov A,X	;restore result
	jnz .looptest					;until status is good
   
continue2:
	;Set CPU clock back to original settings used in project
    M8C_SetBank1
	mov A,[tmp]
	mov reg[OSC_CR0],A
    M8C_SetBank0
    
	;Get starting location of memory area
	SDCard_Cmd 0x50,0,0,0,8,0xFF	;CMD16 - Set block size command
	SDCard_GetR1
	SDCard_CmdEnd
	
	SDCard_Cmd 0x51,0,0,0x01,0xC6,0xFF	;CMD17 - Read single block command (address=454)
	call SDCard_CheckReplyFE
	jnz noCARD						;check result
	;Extract a Boot Sector address
	SDCard_ReadBuffM 4,DataAddr
	SDCard_SkipM 4
	SDCard_CmdEnd
	mov [DataAddr+3],[DataAddr+2]		;x 512
	mov [DataAddr+2],[DataAddr+1]
	mov [DataAddr+1],[DataAddr+0]
	mov [DataAddr+0],0
	asl [DataAddr+1]
	rlc [DataAddr+2]
	rlc [DataAddr+3]

	SDCard_Cmd 0x50,0,0,0,64,0xFF	;CMD16 - Set block size command
	SDCard_GetR1
	SDCard_CmdEnd

	SDCard_CmdA 0x51,DataAddr		;Read block command
	call SDCard_CheckReplyFE
	jnz noCARD						;check result
	;-------- get SD disk information --------
	SDCard_SkipM 11						;skip bytes
	;check Number of bytes per sector = 512
	call SDCard_ReadByte
	cmp A,0
jnz notFAT
	call SDCard_ReadByte
	cmp A,2
jnz notFAT
	;get number of sectors per cluster
	call SDCard_ReadByte
	mov [Sectors_per_cluster],A
	;get number of Reservered sectors x 512 (and calculate FAT address)
	mov [FatAddr+0],0
	call SDCard_ReadByte
	mov [FatAddr+1],A
	call SDCard_ReadByte
	mov [FatAddr+2],A
	mov [FatAddr+3],0
	asl [FatAddr+1]
	rlc [FatAddr+2]
	rlc [FatAddr+3]
	;+ Boot Sector address -> FAT address
	SDCard_Add  FatAddr,DataAddr
	;get number of FAT tables
	call SDCard_ReadByte
	mov [N_of_FAT_tables],A
	cmp A,0	;check =0
jz notFAT
	;get number of root directory enries
	call SDCard_ReadByte
	mov [N_of_dir_enries+0],A
	call SDCard_ReadByte
	mov [N_of_dir_enries+1],A
	SDCard_SkipM 3						;skip bytes
	;Calculate the start address of the Root Directory (sectors per FAT * # of FATs * 512 + FAT start address)
	;get number of sectors per FAT x 512
	call SDCard_ReadByte
	mov [N_of_sectors_per_FAT+0],A
	call SDCard_ReadByte
	mov [N_of_sectors_per_FAT+1],A
	mov [N_of_sectors_per_FAT+2],0
	asl [N_of_sectors_per_FAT+0]
	rlc [N_of_sectors_per_FAT+1]
	rlc [N_of_sectors_per_FAT+2]
	SDCard_Move DirAddr,FatAddr
	loop_calc:
		mov A,[N_of_sectors_per_FAT+0]
		add [DirAddr+1],A
		mov A,[N_of_sectors_per_FAT+1]
		adc [DirAddr+2],A
		mov A,[N_of_sectors_per_FAT+2]
		adc [DirAddr+3],A
		dec [N_of_FAT_tables]
		jnz loop_calc
	;Calculate start of data area (number of root directory enries (usually 512) * entry size(32) + DIR start address
	mov [DataAddr+0],[N_of_dir_enries+0]
	mov [DataAddr+1],[N_of_dir_enries+1]
	mov [DataAddr+2],0
	mov [DataAddr+3],0
	mov [cnt],5
	loop_mul32:
		asl [DataAddr+0]
		rlc [DataAddr+1]
		rlc [DataAddr+2]
		rlc [DataAddr+3]
		dec [cnt]
		jnz loop_mul32
	SDCard_Add  DataAddr,DirAddr	;+ DirAddr
	SDCard_SkipM 30					;skip bytes
	;Double check FAT type info
	call SDCard_ReadByte
	cmp A,'F'
jnz notFAT
	call SDCard_ReadByte
	cmp A,'A'
jnz notFAT
	call SDCard_ReadByte
	cmp A,'T'
jnz notFAT
	call SDCard_ReadByte
	cmp A,'1'
jnz notFAT
	call SDCard_ReadByte
	cmp A,'6'
jnz notFAT
	SDCard_SkipM 4					;skip bytes
	mov [tmp],0
	jmp InitCard_end		;Success!
    
noCARD:
	mov [tmp],1				;if card failed
	jmp InitCard_end
notFAT:
	mov [tmp],2				;if FAT type is missing
InitCard_end:
	SDCard_CmdEnd			;Finish the command cycle
	mov A,[tmp]				;return code

  	ret
.ENDSECTION



;-----------------------------------------------------------------------------
;  FUNCTION: SDCard_FindFile_in_Root
;       Finds a file (or subdir) in the root directory.
;  ARGUMENTS:
;       [ptr]     (2)- address of string 8+3+1 bytes
;       [DirAddr] (2)- start address of directory 
;  RETURNS:
;		A: 0=success, 1=no card, 2=file not found
;       [FatA]  (4) - file start cluster
;       [FSize] (4) - file size
;  Features:
;       1. '*' seeks any char, '?' seeks with char located in altName buffer. 
;       2. 12-th char is an attribute byte: 0x00=file, 0x10=directory, 0x08=label  (Attribute bits: 0=Read Only, 1=Hidden, 2=System, 3=Volume Label, 4=SubDir, 5=Archive)
;-----------------------------------------------------------------------------
.SECTION
 SDCard_FindFile_in_Root::
_SDCard_FindFile_in_Root::
	mov [cntMax+0],[N_of_dir_enries+0]	;load number of entries in root dir
	mov [cntMax+1],[N_of_dir_enries+1]
	SDCard_Move DataA,DirAddr			;initialize DataA pointer (by absolute address)
	jmp SDCard_FindFile_Entry_point

;-----------------------------------------------------------------------------
;  FUNCTION: SDCard_FindFile_in_Cluster
;       Finds a file in cluster.
;  ARGUMENTS:
;       [ptr] (2) - address of string 8+3+1 bytes
;       [ClstN](2)- cluster number
;  RETURNS:
;		A: 0=success, 1=no card, 2=file not found
;       [FClst] (2) - file start cluster
;       [FSize] (4) - file size
;  Features:
;       1. '*' seeks any char, '?' seeks with char located in altName buffer. 
;       2. 12-th char is an attribute byte: 0x00=file, 0x10=directory, 0x08=label  (Attribute bits: 0=Read Only, 1=Hidden, 2=System, 3=Volume Label, 4=SubDir, 5=Archive)
;-----------------------------------------------------------------------------
 SDCard_FindFile_in_Cluster::
_SDCard_FindFile_in_Cluster::
	;load number of entries per cluster (= sect per cluster * 16)
	mov [cntMax+0],[Sectors_per_cluster]
	mov [cntMax+1],0
	mov A,4
	.loop4:
		asl [cntMax+0]
		rlc [cntMax+1]
		dec A
	jnz .loop4	
	call SDCard_Cluster_to_Addr		;cluster number -> absolute address of data
	
SDCard_FindFile_Entry_point:
	SDCard_Cmd 0x50,0,0,0,32,0xFF	;CMD16 - Set block size command
	SDCard_GetR1
	SDCard_CmdEnd

.loop512:
	SDCard_CmdA 0x51,DataA			;Read block command
	call SDCard_CheckReplyFE
	jnz .noCARD						;check result
	mov [cnt],12					;initialize counter - lenght of name + ext + attribute byte
	
	mov X,[ptr+0]					;initialize pointer to name in ROM - [tmp+1]:X
	mov [tmp+1],[ptr+1]
	mov [ptrA],altName				;initialize pointer to alternative chars in RAM

.loop12:
	;read char from SD
	call SDCard_ReadByte			;read byte
	mov [tmp],A						;save current char from SD to [tmp+0]
	;load current char from ROM using pointer [tmp+1]:X
	mov A,[tmp+1]
	romx
	cmp [cnt],1						;is attribute?
	jnz .skip_att					;no
		and [tmp],0x18				;mask unwanted bits
	.skip_att:

	cmp [cnt],12					;1-st char?
	jnz .skip_check					;no
		cmp [tmp],0x00				;record is blank?
		jz .not_found				;yes, that's end of list and file is not found 
		;cmp [tmp],0xFF				;record is blank?
		;jz .not_found				;yes, that's directory end and file is not found 
	.skip_check:
	
	cmp A,'*'						;is "any char"?
	jz .skip_cmp
		cmp A,'?'					;is "alternative char"?
		jnz .skip_alt
			mvi A,[ptrA]			;load char from altName array
		.skip_alt:
		or A,0x20					;make char from ROM lowercase
		or [tmp],0x20				;make char from SD lowercase
		cmp A,[tmp]					;compare chars
		jnz .next_record			;not equal
	.skip_cmp:

	inc X							;increment pointer
	adc [tmp+1],0
	dec [cnt]						;do 12 times
	jnz .loop12
	;now file is found, get other information
	SDCard_SkipM 14
	SDCard_ReadBuffM 2,FClst		;file start cluster
	SDCard_ReadBuffM 4,FSize		;file size
	SDCard_CmdEnd
	mov A,0							;return code
	jmp .end
	
.next_record:
	add [cnt],19
	call SDCard_Skip				;skip uninterested bytes of dir entry 
	SDCard_CmdEnd
	;next record address
	add [DataA+0],32
	adc [DataA+1],0
	adc [DataA+2],0
	adc [DataA+3],0
	;decrement counter
	dec [cntMax+0]
	sbb [cntMax+1],0
	jnz .loop512
	cmp [cntMax+0],0
	jnz .loop512					;repeat until counter > 0
	jmp .not_found2					;else file is not found
.not_found:
	add [cnt],19
	call SDCard_Skip				;skip other bytes of dir entry 
	SDCard_CmdEnd
.not_found2:
	mov A,2							;return code
	jmp .end
.noCARD:
	mov A,1							;return code
.end:
  	ret
.ENDSECTION


;-----------------------------------------------------------------------------
;  FUNCTION: SDCard_Cluster_to_Addr
;       Converts cluster number -> absolute address of data
;                       [ClstN] -> [DataA]
;  ARGUMENTS:
;       [ClstN] (2)- cluster number
;  RETURNS:
;       [DataA] (4) - absolute address of data
;-----------------------------------------------------------------------------
.SECTION
SDCard_Cluster_to_Addr::
	;absolute address = data area address + (cluster number - 2) * sectors_per_cluster * 512
	sub [ClstN+0],2							;ClstN = ClstN -2 
	sbb [ClstN+1],0
	SDCard_mul	DataA,ClstN,Sectors_per_cluster		;DataA = ClstN * Sectors_per_cluster
	mov [DataA+3],[DataA+2]					;DataA * 512
	mov [DataA+2],[DataA+1]
	mov [DataA+1],[DataA+0]
	mov [DataA+0],0
	asl [DataA+1]
	rlc [DataA+2]
	rlc [DataA+3]
	SDCard_Add  DataA,DataAddr				;DataA =DataA + DataAddr 
	add [ClstN+0],2							;restore cluster number
	adc [ClstN+1],0
  	ret
.ENDSECTION

;-----------------------------------------------------------------------------
;  FUNCTION: SDCard_NextCluster
;       Reads a data in FAT area to define next cluster in chain
;  ARGUMENTS:
;       [ClstN] (2) - cluster number, is replaced by new
;  RETURNS:
;		A: 0=success, 1=no card
;-----------------------------------------------------------------------------
.SECTION
SDCard_NextCluster::
	;define next cluster (2 bytes from address: FAT area address + (cluster number) * 2)
	mov [DataA+0],[ClstN+0]	;DataA <- ClstN
	mov [DataA+1],[ClstN+1]
	mov [DataA+2],0
	mov [DataA+3],0
		asl [DataA+0]				;DataA = DataA * 2
		rlc [DataA+1]
		rlc [DataA+2]
	SDCard_Add	DataA,FatAddr		;DataA = DataA + FatAddr

	SDCard_Cmd 0x50,0,0,0,2,0xFF	;CMD16 - Set block size command
	SDCard_GetR1
	SDCard_CmdEnd

	SDCard_CmdA 0x51,DataA			;Read block command
	call SDCard_CheckReplyFE
	jnz .noCARD						;check result
	SDCard_ReadBuffM 2,ClstN
	SDCard_CmdEnd
	mov A,0							;return code
  	ret
.noCARD:
	mov A,1							;return code
	ret
.ENDSECTION

;-----------------------------------------------------------------------------
;  FUNCTION: SDCard_ReadDataBuff
;       Reads all data by buffer 32 bytes
;  ARGUMENTS:
;       [ClstN] (2)- cluster number
;  RETURNS:
;		A: 0=success, 1=no card
;-----------------------------------------------------------------------------
.SECTION
SDCard_ReadDataBuff::
	mov [ClstN+0],[FClst+0]				;load start cluster number
	mov [ClstN+1],[FClst+1]
	SDCard_Move FSizeCnt,FSize			;initialize size counter

SDCard_ReadData_loop:
	call SDCard_Cluster_to_Addr		;convert [ClstN] -> [DataA] (absolute address)
	SDCard_Cmd 0x50,0,0,0,BUFFERSIZE,0xFF	;CMD16 - Set block size command
	SDCard_GetR1
	SDCard_CmdEnd

	mov [tmp+1],[Sectors_per_cluster]	;initialize sector counter

SDCard_ReadData_loop1:
	mov [tmp+0],(512/BUFFERSIZE)		;initialize shunk counter

SDCard_ReadData_loop2:
	mov A,[FSizeCnt+0]				;check size counter (=0?)
	or  A,[FSizeCnt+1]
	or  A,[FSizeCnt+2]
	or  A,[FSizeCnt+3]
	jz SDCard_ReadData_end			;if size counter =0

	sub [FSizeCnt+0],BUFFERSIZE		;decrement size counter
	sbb [FSizeCnt+1],0
	sbb [FSizeCnt+2],0
	sbb [FSizeCnt+3],0
 
	SDCard_CmdA 0x51,DataA			;Read block command
	call SDCard_CheckReplyFE
	jnz SDCard_ReadData_end_noCARD	;check result
	;   -   SDCard_ReadBuffM BUFFERSIZE,(buff+0)
	mov [cnt],BUFFERSIZE
	mov X,[ptrW]					;load pointer
	call SDCard_ReadBuff
	SDCard_CmdEnd

	;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
	;; now data in buff is ready. [ptrW] points to 1-st of BUFFERSIZE bytes
		;if bit 7 of [FSizeCnt+3] is 1,  [FSizeCnt+0]+BUFFERSIZE is a number of valid bytes in buff
		mov A,[FSizeCnt+0]
		add A,BUFFERSIZE
		tst [FSizeCnt+3],0x80		;check bit 7
		jnz .skip					;if not all bytes of buff are valid
			mov A,BUFFERSIZE		;else flag A = number of valid bytes
		.skip:
	
	;increment ptrW pointer 
	add [ptrW],A;BUFFERSIZE
	cmp [ptrW],buff+BUFFERSIZE*BUFFNUMBER
	jc .skip_init
		sub [ptrW],BUFFERSIZE*BUFFNUMBER
	.skip_init:


	;; check for free space between ptrW and ptrR
	.wait_free_space:
	mov A,[ptrR]
	sub A,[ptrW]
	jnc .skip_corr
		add A,BUFFERSIZE*BUFFNUMBER
	.skip_corr:
	cmp A,BUFFERSIZE+1
	jc .wait_free_space

	;call SDCard_Callback_Procedure
	;;
	;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
	
	tst [FSizeCnt+3],0x80			;check bit 7
	jnz SDCard_ReadData_end			;if size counter <0

	add [DataA+0],BUFFERSIZE		;next data block address
	adc [DataA+1],0
	adc [DataA+2],0
	adc [DataA+3],0

	dec [tmp+0]
	jnz SDCard_ReadData_loop2		;do 512/BUFFERSIZE times
	dec [tmp+1]
	jnz SDCard_ReadData_loop1		;do Sectors_per_cluster times

	call SDCard_NextCluster
	cmp A,1
	jz SDCard_ReadData_end_noCARD	;check result
	;check cluster number (end of chain)
	cmp [ClstN+1],0xFF
	jnz SDCard_ReadData_loop
	cmp [ClstN+0],0xF8
	jc  SDCard_ReadData_loop		;do while not end of chain (ClstN < FFF8)
SDCard_ReadData_end:
	mov A,0							;return code
	ret
SDCard_ReadData_end_noCARD:
	mov A,1							;return code
	ret
.ENDSECTION


;-----------------------------------------------------------------------------
;  FUNCTION: SDCard_Run
;       Run all procedures to read file (using path, name from ROM + alternative chars from RAM)
;  ARGUMENTS:
;       see code
;  RETURNS:
;		A: 0=success, 1=no card, 2=file not found
;-----------------------------------------------------------------------------
.SECTION
 SDCard_Run::
_SDCard_Run::
;-----------------------------------------------	
stage0:	;Initialization of SD
;-----------------------------------------------	
	lcall SDCard_InitCard
	cmp A,0
	jz stage1				;if success
	mov A,1					;if no card or FAT16
	jmp return_code
;-----------------------------------------------	
stage1:	;Find File(Directory) in RootDir
;-----------------------------------------------	
	mov [ptr+0],<name1				;load name pointer
	mov [ptr+1],>name1
	lcall SDCard_FindFile_in_Root
	cmp A,0
	jnz return_code					;if no card or no file
;-----------------------------------------------	
stage2:	;Find in SubDir
;-----------------------------------------------	
	mov [ClstN+0],[FClst+0]			;load start cluster number
	mov [ClstN+1],[FClst+1]
	mov [ptr+0],<name2				;load name pointer
	mov [ptr+1],>name2
stage2_loop:
	lcall SDCard_FindFile_in_Cluster
	cmp A,0
	jz stage3						;if file is found
	cmp A,1
	jz return_code					;if no card
	;file not found - go to next cluster
	call SDCard_NextCluster			;Reads a data in FAT area to define next cluster in chain
	;check cluster number (end of chain)
	cmp [ClstN+0],0xFF
	jnz stage2_loop
	cmp [ClstN+1],0xFF
	jnz stage2_loop
	;end of chain is reached, file is not found
	mov A,2
	jmp return_code
;-----------------------------------------------	
stage3:	;Read Data
;-----------------------------------------------	
	call SDCard_ReadDataBuff
	cmp A,0
	jz	stage4		;if success
	mov A,1
	jmp return_code
;-----------------------------------------------	
stage4:	;Data processings if required
;-----------------------------------------------	
	
	;    
;	toLCD_Position  0, 8    
	;toLCD_9_99 BSAddr ;Boot Sector address
;	toLCD_9_99 FatAddr ;Address of FAT area
;	toLCD_Position  1, 0    
;	toLCD_9_99 DirAddr ;Address of Dir area
;	toLCD_Position  1, 8    
;	toLCD_9_99 DataAddr ;Address of Data area
;	jmp t

	mov A,0	;return code
return_code:
	ret	
.ENDSECTION

;-----------------------------------------------------------------------------
;  FUNCTION: SDCard_Callback_Procedure
;       Callback procedure that is run from SDCard_ReadDataBuff
;  ARGUMENTS:
;       A - number of valid bytes
;       [buff] - data
;-----------------------------------------------------------------------------
.SECTION
 SDCard_Callback_Procedure::
_SDCard_Callback_Procedure::

	;example
	;include "to_LCD.inc" 
	;push A ;save number of valid bytes 
	;toLCD_Position  1, 0    
	;mov [buff+16],0;zero-terminated string
	;toLCD_PrString  buff
	
	;toLCD_Position  0, 10    
	;pop A ;restore number of valid bytes
	;lcall LCD_PrHexByte
	ret
.ENDSECTION

;-----------------------------------------------------------------------------
;  DATA section - Name array
;  Note: * = any char, ? = char is taken from the alternative array (from RAM)
;-----------------------------------------------------------------------------
.LITERAL
;   1   2   3   4   5   6   7   8     1   2   3  Attr: 0x00=file, 0x10=directory, 0x08=label
name1:
DB 'M','U','Z','D','A','T','A',' ',  '*','*','*',0x10
;DB 'M','U','Z','*','*','*','*',' ',  '*','*','*',0x10
;DB 'D','C','I','M',' ',' ',' ',' ',  '*','*','*',0x10
;DB '0','0','1','L','a','m','b','a',  't','x','t',0x00

;   1   2   3   4   5   6   7   8     1   2   3  Attr
name2:
;DB '0','0','1','L','a','m','b','a',  't','x','t',0x00
DB '?','?','?','*','*','*','*','*',  'b','o','r',0x00
;DB '0','0','7','*','*','*','*','*',  'b','o','r',0x00
;DB '*','*','*','N','*','*','*','*',  '*','*','*',0x10
;DB '1','0','0','N','I','K','O','N',  '*','*','*',0x10
;DB 'T','e','x','t','1',' ',' ',' ',  't','x','t',0x00
.ENDLITERAL


; End of File SDCard.asm
